PetInfoFu = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0",
                                           "AceConsole-2.0",
                                           "AceDB-2.0",
                                           "AceEvent-2.0"
                                          )

 -- variables
local L          = AceLibrary("AceLocale-2.2"):new("PetInfoFu")
local BS         = AceLibrary("Babble-Spell-2.2")
local BST        = AceLibrary("Babble-SpellTree-2.2")
local crayon     = AceLibrary("Crayon-2.0")
local dewdrop    = AceLibrary("Dewdrop-2.0")
local tablet     = AceLibrary("Tablet-2.0")

-- hunter specific variables
local prevHappiness = 0                 -- tracks happiness level (0-3)
-- happiness tracking
local DECAY_RATE = 5/8
local estimatedPetHappiness = 175       -- set to mid happiness
local isFeedingPet = false
-- level timers
local timerZero = 0
local petXPZero = 0
local prevXPToLevel = 0
local lastKillXP = 0
local prevTP = 0
local prevPetLevel = 0

-- warlock specific variables
local demonicKnowledgeRank = 0
local demonicKnowledgeTalentTab = 0
local demonicKnowledgeTalentIndex = 0

-- properties
PetInfoFu.clickableTooltip = true
PetInfoFu.canHideText      = true
PetInfoFu.defaultPosition  = "RIGHT"
PetInfoFu.hasIcon          = true
PetInfoFu.hasNoColor       = false

-- register database stuff
PetInfoFu:RegisterDB("PetInfoFuDB")
PetInfoFu:RegisterDefaults("account", {
                                        stables = {},
                                      }
                          )
                          
PetInfoFu:RegisterDefaults("char", {
                                    -- flags
                                    showPetName = true,
                                    showPetLevel = true,
                                    showPetXP = 0,
                                    showPetXPOnBar = false,
                                    showAttributes = true,
                                    showCombatStats = true,
                                    showSkills = true,
                                    showStableMapNotes = false,
                                    showStabledPets = false,
                                    showLoyalty = false,
                                    trackPetHistory = false,
                                    trackPetHappiness = false,
                                    trackPetSkills = true,
                                    -- sound play
                                    playSoundOnGainTP = false,
                                    playSoundOnSkillTrainable = false,
                                    hasPlayedSoundOnGainTP = false,
                                    hasPlayedSoundOnSkillTrainable =false,
                                    -- frame locations
                                    petInfoFramePoint = { point = "CENTER", x = 0, y = 0},
                                    petExperienceFramePoint = { point = "CENTER", x = 0, y = 0},
                                    -- lists
                                    hidden = {},
                                    stabledPets = {},       -- stores currently stabled pets
                                    petHistory = {},        -- stores list of all pets ever stabled
                                    beastTraining = {},     -- stores hunter known training spells
                                    petInfo = {},           -- stores pet skills and similar info
                                    netheredDemons = {},    -- stores info about all demons
                                   } 
                         )
PetInfoFu:RegisterChatCommand(L["CMD_OPTIONS"], options)

-- menu methods

function PetInfoFu:IsShowingPetName()
	return self.db.char.showPetName
end

function PetInfoFu:ToggleIsShowingPetName()
	self.db.char.showPetName = not self.db.char.showPetName
    if not (self.db.char.showPetName or self.db.char.showPetLevel) then self:ShowIcon() end
	self:UpdateText()
	return self.db.char.showPetName
end

function PetInfoFu:IsShowingPetLevel()
	return self.db.char.showPetLevel
end

function PetInfoFu:ToggleIsShowingPetLevel()
	self.db.char.showPetLevel = not self.db.char.showPetLevel
    if not (self.db.char.showPetName or self.db.char.showPetLevel) then self:ShowIcon() end
	self:UpdateText()
	return self.db.char.showPetLevel
end

function PetInfoFu:IsShowingXPByAmount()
	return self.db.char.showPetXP == 1
end

function PetInfoFu:ToggleIsShowingXPByAmount()
	if self.db.char.showPetXP == 1 then 
        self.db.char.showPetXP = 0
    else
        self.db.char.showPetXP = 1
    end
	self:UpdateText()
	return self.db.char.showPetXP == 1
end

function PetInfoFu:IsShowingXPByPercentage()
	return self.db.char.showPetXP == 2
end

function PetInfoFu:ToggleIsShowingXPByPercentage()
	if self.db.char.showPetXP == 2 then 
        self.db.char.showPetXP = 0
    else
        self.db.char.showPetXP = 2
    end
	self:UpdateText()
	return self.db.char.showPetXP == 2
end

function PetInfoFu:IsShowingXPByTime()
	return self.db.char.showPetXP == 3
end

function PetInfoFu:ToggleIsShowingXPByTime()
	if self.db.char.showPetXP == 3 then 
        self.db.char.showPetXP = 0
    else
        self.db.char.showPetXP = 3
        self:ResetXPTimer()
    end
	self:UpdateText()
	return self.db.char.showPetXP == 3
end

function PetInfoFu:IsShowingXPByKill()
	return self.db.char.showPetXP == 4
end

function PetInfoFu:ToggleIsShowingXPByKill()
	if self.db.char.showPetXP == 4 then 
        self.db.char.showPetXP = 0
    else
        self.db.char.showPetXP = 4
    end
	self:UpdateText()
	return self.db.char.showPetXP == 4
end

function PetInfoFu:IsShowingXPOnBar()
	return self.db.char.showPetXPOnBar
end

function PetInfoFu:ToggleIsShowingXPOnBar()
	self.db.char.showPetXPOnBar = not self.db.char.showPetXPOnBar 
	self:UpdateText()
	return self.db.char.showPetXPOnBar
end
function PetInfoFu:IsShowingAttributes()
	return self.db.char.showAttributes
end

function PetInfoFu:ToggleIsShowingAttributes()
	self.db.char.showAttributes = not self.db.char.showAttributes
	self:UpdateTooltip()
	return self.db.char.showAttributes
end

function PetInfoFu:IsShowingCombatStats()
	return self.db.char.showCombatStats
end

function PetInfoFu:ToggleIsShowingCombatStats()
	self.db.char.showCombatStats = not self.db.char.showCombatStats
	self:UpdateTooltip()
	return self.db.char.showCombatStats
end

function PetInfoFu:IsShowingSkills()
	return self.db.char.showSkills
end

function PetInfoFu:ToggleIsShowingSkills()
	self.db.char.showSkills = not self.db.char.showSkills
	self:UpdateTooltip()
	return self.db.char.showSkills
end

function PetInfoFu:IsShowingStableMapNotes()
	return self.db.char.showStableMapNotes
end

function PetInfoFu:ToggleIsShowingStableMapNotes()
	self.db.char.showStableMapNotes = not self.db.char.showStableMapNotes
	self:UpdateTooltip()
	return self.db.char.showStableMapNotes
end

function PetInfoFu:IsShowingStabledPets()
	return self.db.char.showStabledPets
end

function PetInfoFu:ToggleIsShowingStabledPets()
	self.db.char.showStabledPets = not self.db.char.showStabledPets
	self:UpdateTooltip()
	return self.db.char.showStabledPets
end

function PetInfoFu:IsShowingLoyalty()
	return self.db.char.showLoyalty
end

function PetInfoFu:ToggleIsShowingLoyalty()
	self.db.char.showLoyalty = not self.db.char.showLoyalty
	self:UpdateTooltip()
	return self.db.char.showLoyalty
end

function PetInfoFu:IsTrackingPetHistory()
	return self.db.char.trackPetHistory
end

function PetInfoFu:ToggleIsTrackingPetHistory()
	self.db.char.trackPetHistory = not self.db.char.trackPetHistory
	self:UpdateText()
	return self.db.char.trackPetHistory
end

function PetInfoFu:IsTrackingPetHappiness()
	return self.db.char.trackPetHappiness
end

function PetInfoFu:ToggleIsTrackingPetHappiness()
	self.db.char.trackPetHappiness = not self.db.char.trackPetHappiness
	self:UpdateText()
	return self.db.char.trackPetHappiness
end

function PetInfoFu:ResetXPTimer()
	timerZero = time()
    _, _, petXPZero = self:PetInfo_GetPetExperience()
end

function PetInfoFu:IsPlayingSoundOnGainTP()
	return self.db.char.playSoundOnGainTP
end

function PetInfoFu:TogglePlaySoundOnGainTP()
	self.db.char.playSoundOnGainTP = not self.db.char.playSoundOnGainTP
	self:Update()
	return self.db.char.playSoundOnGainTP
end
function PetInfoFu:IsPlayingSoundOnSkillTrainable()
	return self.db.char.playSoundOnSkillTrainable
end

function PetInfoFu:TogglePlaySoundOnSkillTrainable()
	self.db.char.playSoundOnSkillTrainable = not self.db.char.playSoundOnSkillTrainable
	self:Update()
	return self.db.char.playSoundOnSkillTrainable
end
                                          
-- init/exit functions

function PetInfoFu:OnInitialize()
  --nothing here
end

function PetInfoFu:OnEnable()
    -- update
    self:RegisterEvent("PLAYER_ENTERING_WORLD","Update")
	self:RegisterEvent("PLAYER_LEVEL_UP","Update")
    ---- seems not to catch a change in demon, possibly obsolete
    ---- self:RegisterEvent("PLAYER_PET_CHANGED","Update")
    -- UNIT_PET: arg1 = who, if arg1=='player' then it's our pet
    self:RegisterEvent("UNIT_PET")
    
    -- local event handlers
    if self:IsHunter() then
        self:RegisterEvent("CRAFT_CLOSE")
        self:RegisterEvent("PET_STABLE_CLOSED")
        self:RegisterEvent("PET_STABLE_SHOW")
    	self:RegisterEvent("UNIT_HAPPINESS")
	    self:RegisterEvent("UNIT_PET_EXPERIENCE")
    	self:RegisterEvent("UNIT_PET_TRAINING_POINTS")
        -- happiness tracking
        self:RegisterEvent("CHAT_MSG_SPELL_AURA_GONE_SELF")        
        self:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS")
        self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
        --      
        self:ScheduleRepeatingEvent(self.name, self.UpdateText, 60, self)         
    end
    if self:IsWarlock() then
        self:GetDemonicKnowledgeRank()
        self:RegisterEvent("CHARACTER_POINTS_CHANGED")
    end
    
    self:SetIcon(BS:GetSpellIcon("Train Beast"))
    
    -- mapnotes
 	if Cartographer_Notes then
    	Cartographer_Notes:RegisterIcon("Stable", {text = L["Stable"], 
                                                   path = "Interface\\AddOns\\FuBar_PetInfoFu\\icons\\stable.tga"
                                                }
                                        )
        Cartographer_Notes:RegisterNotesDatabase('PetInfoFu', self.db.account.stables, self)
    end
    
    -- set timers and other global variables
    self:ResetXPTimer()
    self:CheckForNextLevelSkill()
end

function PetInfoFu:OnDisable()
    if Cartographer_Notes then
		Cartographer_Notes:UnregisterIcon("Stable")  
        Cartographer_Notes:UnregisterNotesDatabase('PetInfoFu')
    end
    self:UnregisterAllEvents()
end

-- events (sorted alphabetically)
function PetInfoFu:CHARACTER_POINTS_CHANGED()
    self:GetDemonicKnowledgeRank()
end

function PetInfoFu:CHAT_MSG_SPELL_AURA_GONE_SELF()
    -- catch the Feed Pet Effect fades
    if string.find(arg1, BS["Feed Pet Effect"]) ~= nil then 
        -- DEFAULT_CHAT_FRAME:AddMessage('PetInfoFu: AURA GONE '..crayon:Orange(arg1))
        isFeedingPet = false
    end
end

function PetInfoFu:CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS() 
    if string.find(arg1, BS["Feed Pet Effect"]) ~= nil then 
        --DEFAULT_CHAT_FRAME:AddMessage('PetInfoFu: SELF BUFF '..crayon:Orange(arg1))
        i = string.find(arg1, L["Happiness"])
        if i ~= nil then 
            --                 i
            -- Lazuli gains 35 Happiness from Pet Feed Effect
            local tmpHappiness = string.sub(arg1,i-3,i-1)
            self:CalculateEstimatedHappiness(tonumber(tmpHappiness)) -- coerce to numerical value
            self:UpdateText()
        end
    end 
end

function PetInfoFu:CRAFT_CLOSE()
    -- make sure we don't get enchanting items in here
    if GetCraftSkillLine(1) == BS["Beast Training"] then
    -- write pet skills known by hunter to database
    for i = 1, GetNumCrafts() do
        local craftName, craftSubSpellName, _, _, _, _, requiredLevel = GetCraftInfo(i)
        if craftName ~= nil then
            if self.db.char.beastTraining[craftName] == nil then
                self.db.char.beastTraining[craftName] = {}                        
            end
            if self.db.char.beastTraining[craftName][craftSubSpellName] == nil then 
                self.db.char.beastTraining[craftName][craftSubSpellName] = {
                    RequiredLevel = requiredLevel
                }
            end
        end
    end
end    
end    

function PetInfoFu:PET_STABLE_CLOSED()
    self:GetStabledPets()
    if self.db.char.trackPetHistory then
        self:TrackPetHistory()
    end
    if Cartographer_Notes and 
       self.db.char.showStableMapNotes then
        self:AddStableMapNote()
    end
end

function PetInfoFu:PET_STABLE_SHOW()
    if self.db.char.trackPetSkills then self:StoreCurrentPet() end
end

function PetInfoFu:UNIT_HAPPINESS()
    -- update happiness state (unhappy, content, happy)
    local happinessLevel = GetPetHappiness()
    if happinessLevel ~= nil then        
        if happinessLevel ~= prevHappiness then        
            -- calibrate numerical happiness
            if prevHappiness ~= 0 then 
                if happinessLevel > prevHappiness then
                    estimatedPetHappiness = 0      -- reset to start of higher level
    end
                if happinessLevel < prevHappiness then
                    estimatedPetHappiness = 350    -- reset to top of lower level
    end
        end
            -- set tracking variable and update everything                
            prevHappiness = happinessLevel            
            self:Update()
    end 
    -- update the numerical happiness value
    if not isFeedingPet then
            self:CalculateEstimatedHappiness(-DECAY_RATE)
        self:UpdateText()        
    end
    end
end    
        
-- add code to reset pet specific globals in here
function PetInfoFu:UNIT_PET()
    if arg1 == 'player' then 
        self:Update()
    end
end

function PetInfoFu:UNIT_PET_EXPERIENCE()
    if self.db.char.showPetXP == 4 then
        local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
        if prevXPToLevel == 0 then prevXPToLevel = toLevelXP end
        lastKillXP = prevXPToLevel - toLevelXP
        if lastKillXP < 0 then lastKillXP = toLevelXP end
        prevXPToLevel = toLevelXP      
    end
    self:Update()
    self:UpdatePetExperienceFrame()
end

-- add functionality for training points warning in here
function PetInfoFu:UNIT_PET_TRAINING_POINTS()
    if self:IsHunter() then
        self:CheckForTPGain()
        self:CheckForNextLevelSkill() 
        self:Update()
    end
end

function PetInfoFu:UNIT_SPELLCAST_SUCCEEDED()
-- arg1: unit casting the spell 
-- arg2: spell name 
-- arg3 : spell rank 
    if arg2 == BS["Feed Pet"] then -- feeding pet, start tracking 
        isFeedingPet = true
    end
    if arg2 == BS["Dismiss Pet"] then -- dismiss pet, substract 50
        self:CalculateEstimatedHappiness(-50)
    end
    if arg2 == BS["Improved Revive Pet"] or -- pet has died and is ressed
        arg2 == BS["Revive Pet"] then
        -- if this is 350, the counter can stay where it is, as colour will change
        -- self:CalculateEstimatedHappiness(-350)
    end
end;

-- frame functions

function PetInfoFu:UpdatePetExperienceFrame()
	if self.petExperienceFrame then
        local text = ''
        -- show XP
        local petLevel = UnitLevel("pet")
        local toonLevel = UnitLevel("player")        
        if petLevel ~= toonLevel then       
            if self.db.char.showPetXP == 1 then                                             -- show by amount
                local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
                text = L["To go: "]..format(L["PI_FORMAT"], toLevelXP)
            elseif self.db.char.showPetXP == 2 then                                         -- show by percent 
                local _, totalXP, toLevelXP = self:PetInfo_GetPetExperience()
                local toLevelXPPercent = toLevelXP / totalXP * 100                   
                text = L["To go: "]..format(L["PI_PERCENT_FORMAT"],toLevelXPPercent)
            elseif self.db.char.showPetXP == 3 then                                         -- show by time
                local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
                text = L["To go: "]..self:GetTimeToLevel(toLevelXP)
            elseif self.db.char.showPetXP == 4 then                                         -- show by kill
                if lastKillXP > 0 then 
                    local _, _, toLevelXP = self:PetInfo_GetPetExperience()
                    text = L["To go: "]..self:GetKillsToLevel(toLevelXP)
                end
            end
        end
        self.petExperienceFrame.petText:SetText(text)            
    end
end

function PetInfoFu:PetExperienceFrame()
    -- hide other frames, destroy if same frame is called again
	if self.petExperienceFrame then 
		self.petExperienceFrame:Hide()
		if self.petExperienceFrame.name ==UnitName("pet") then
			self.petExperienceFrame = nil
			return
		end
	end	
    
    local text = 'Pet Experience'
    
    -- create frame
	local frame = CreateFrame("Frame", "PetInfoFuExperienceFrame", UIParent)
	frame:SetPoint(self.db.char.petExperienceFramePoint.point, self.db.char.petExperienceFramePoint.x, self.db.char.petExperienceFramePoint.y)
	frame:SetFrameStrata("TOOLTIP")
	frame:SetFrameLevel(99)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", function()
		this:StartMoving()
	end)
	frame:SetScript("OnDragStop", function()
		this:StopMovingOrSizing()
		local point,relativeTo,relativePoint,xOfs,yOfs = this:GetPoint()
		self.db.char.petExperienceFramePoint.point = point
		self.db.char.petExperienceFramePoint.x = xOfs
		self.db.char.petExperienceFramePoint.y = yOfs
	end)

	frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4},
	})
	frame:SetBackdropBorderColor(.5, .5, .5, 1)
	frame:SetBackdropColor(0, 0, 0, .3)
    
	-- create pet icon
    local petIcon = frame:CreateTexture(nil, "ARTWORK")
    frame.petIcon = petIcon
    petIcon:SetTexture(GetPetIcon())
    petIcon:SetWidth(26)
    petIcon:SetHeight(26)
    petIcon:SetPoint("TOPRIGHT", frame, "TOPLEFT", 1, -3)

	
	-- create close button
	local closeButton = CreateFrame("Button", "PetInfoFuExpCloseButton", frame)
	closeButton:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0)
	closeButton:SetWidth(30)
	closeButton:SetHeight(30)
	closeButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
	closeButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down")
	closeButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight")
	closeButton:SetScript("OnClick", function()
		frame:Hide()
	end)

    -- set actual tooltip text
	local petText = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	frame.petText = petText
	petText:SetText(text)
	petText:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -10)
	petText:SetWidth(140)
	petText:SetJustifyH("LEFT")
	petText:SetJustifyV("TOP")

    -- update text
    self:UpdatePetExperienceFrame()
    
    -- activate frame
	frame.name = UnitName("pet")
	self.petExperienceFrame = frame
	frame:SetWidth(petText:GetWidth() + 20)
	frame:SetHeight(petText:GetHeight() + 20)
	frame:Show()    
end

function PetInfoFu:PetInfoFrame(petInfo)
    -- with kind thanks to TrainerFu for the basic routine
    
    -- hide other frames, destroy if same frame is called again
	if self.petInfoFrame then 
		self.petInfoFrame:Hide()
		if self.petInfoFrame.name == petInfo.PetName then
			self.petInfoFrame = nil
			return
		end
	end	
   
    local text = crayon:Green(petInfo.PetName..' ('..petInfo.PetLevel..')\n')..petInfo.PetType
    if petInfo.PetLoyalty ~= nil then text = text..' - '..petInfo.PetLoyalty end   
	
    local petSaveKey = petInfo.PetType.."|"..petInfo.PetName
    -- if available add list of pet skills
    if self.db.char.petInfo[petSaveKey] ~= nil then
        if self.db.char.petInfo[petSaveKey]['petSkills'] ~= nil then
            text = text..'\n'
            -- create a sorted table of pet skills
            sortedPetSkills = {}
            for k,v in pairs(self.db.char.petInfo[petSaveKey]['petSkills']) do
                -- DEFAULT_CHAT_FRAME:AddMessage('PetInfoFu: Adding'..crayon:Red(k)..' '..v.SpellRank)
                if v.SpellRank == '' then
                    table.insert(sortedPetSkills, k)                
                else
                    table.insert(sortedPetSkills, k..' - '..v.SpellRank)
                end
            end
            table.sort(sortedPetSkills, function (a, b) return a < b end)
            -- list sorted pet skills
            for i = 1, #sortedPetSkills do  
                text = text..'\n'..sortedPetSkills[i]
            end
         end
    end
	
    -- create frame
	local frame = CreateFrame("Frame", "PetInfoFuPetFrame", UIParent)
	frame:SetPoint(self.db.char.petInfoFramePoint.point, self.db.char.petInfoFramePoint.x, self.db.char.petInfoFramePoint.y)
	frame:SetFrameStrata("TOOLTIP")
	frame:SetFrameLevel(99)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", function()
		this:StartMoving()
	end)
	frame:SetScript("OnDragStop", function()
		this:StopMovingOrSizing()
		local point,relativeTo,relativePoint,xOfs,yOfs = this:GetPoint()
		self.db.char.petInfoFramePoint.point = point
		self.db.char.petInfoFramePoint.x = xOfs
		self.db.char.petInfoFramePoint.y = yOfs
	end)

	frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4},
	})
	frame:SetBackdropBorderColor(.5, .5, .5, 1)
	frame:SetBackdropColor(0, 0, 0, .3)
    
	-- create pet icon
    if self.db.char.petInfo[petSaveKey].PetIcon ~= nil then        
        local petIcon = frame:CreateTexture(nil, "ARTWORK")
        frame.petIcon = petIcon
        petIcon:SetTexture(self.db.char.petInfo[petSaveKey].PetIcon)
        petIcon:SetWidth(40)
        petIcon:SetHeight(40)
        petIcon:SetPoint("TOPRIGHT", frame, "TOPLEFT", 1, -3)
    end
	
	-- create close button
	local closeButton = CreateFrame("Button", "PetInfoFuCloseButton", frame)
	closeButton:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0)
	closeButton:SetWidth(30)
	closeButton:SetHeight(30)
	closeButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
	closeButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down")
	closeButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight")
	closeButton:SetScript("OnClick", function()
		frame:Hide()
	end)

    -- set actual tooltip text
	local petText = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	frame.petText = petText
	petText:SetText(text)
	petText:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -10)
	petText:SetWidth(280)
	petText:SetJustifyH("LEFT")
	petText:SetJustifyV("TOP")

    -- activate frame
	frame.name = petInfo.petName
	self.petInfoFrame = frame
	frame:SetWidth(petText:GetWidth() + 20)
	frame:SetHeight(petText:GetHeight() + 20)
	frame:Show()    
end

-- functions 

function PetInfoFu:StoreCurrentPet()
    local petName = UnitName("pet")
    
    if petName ~= nil and 
        petName ~= UNKNOWNOBJECT then        
        -- DEFAULT_CHAT_FRAME:AddMessage('PetInfoFu: Storing '..crayon:Red(petName))    
        local petType = UnitCreatureFamily("pet")
        local petSaveKey = petType.."|"..petName    
        -- init list item if not exists
        if self.db.char.petInfo[petSaveKey] == nil then
            self.db.char.petInfo[petSaveKey] = {}        
        end
        self.db.char.petInfo[petSaveKey] = {
            PetName = petName,
            PetType = petType,
            PetLevel = UnitLevel("pet"),
            PetHappiness = GetPetHappiness(),
            PetIcon = GetPetIcon()
        }
        -- list pet skills, reset curent list
        self.db.char.petInfo[petSaveKey]['petSkills'] = {}
        local i = 1
        while true do
            local spellName, spellRank = GetSpellName(i, BOOKTYPE_PET)
            if not spellName then
                do break end
            end
            self.db.char.petInfo[petSaveKey]['petSkills'][spellName]  = {
                SpellRank = spellRank
            }
            i = i + 1
        end
    end
end

function PetInfoFu:CalculateEstimatedHappiness(delta)
    if delta == nil then delta = 0 end  -- catch nil
    if delta == 0 then                  -- pet is filled up
        estimatedPetHappiness = 350
    else    
        estimatedPetHappiness = estimatedPetHappiness + delta
    end
    
    local happinessLevel = GetPetHappiness()   
    if estimatedPetHappiness > 350 then 
        if happinessLevel == 3 then -- happy, then cap at 350
            estimatedPetHappiness  = 350 
        else    
            estimatedPetHappiness = estimatedPetHappiness - 350
        end
    end
    if estimatedPetHappiness < 0 then -- we passed a happiness level reset to 350 for full level
       estimatedPetHappiness = estimatedPetHappiness + 350
    end
    -- DEFAULT_CHAT_FRAME:AddMessage('PetInfoFu: '..crayon:Red('Happiness: ')..format(estimatedPetHappiness)..' ('..format(delta)..')')
end
function PetInfoFu:GetStabledPets()
    self.db.char.stabledPets = nil
    self.db.char.stabledPets = {}
    
    for i = 1, GetNumStableSlots() do  -- i = 0 for current pet
        local _, petName, petLevel, petType, petLoyalty = GetStablePetInfo(i);
        if petName ~= nil then
            -- DEFAULT_CHAT_FRAME:AddMessage('PetInfoFu: Stabling '..crayon:Red(petName)..' '..format(i))         
            self.db.char.stabledPets[i] = {
                PetName = petName,
                PetType = petType,
                PetLoyalty = petLoyalty,
                PetLevel = petLevel
            }
        end
    end
    self:Update()
end

function PetInfoFu:TrackPetHistory()
    -- store each pet only once, with type and name as key
    -- this will get messed up if you don't name your pets 
    -- and stable a new pet of the same type or rename 
    -- a pet with the same name as an existing pet of that type
    for i = 1, GetNumStableSlots() do  -- i = 0 for current pet
        local _, petName, petLevel, petType, petLoyalty = GetStablePetInfo(i);
        if petName ~= nil then
            if self.db.char.petHistory[petType.."|"..petName] == nil then
                self.db.char.petHistory[petType.."|"..petName] = {
                    PetName = petName,
                    PetType = petType,
                    PetLevel = petLevel,
                    HunterLevel = UnitLevel("player")                
                }
            end
        end
    end
    self:Update()
end

function PetInfoFu:GetDemonicKnowledgeRank()   
     if (demonicKnowledgeTalentTab == 0 or demonicKnowledgeTalentIndex == 0) then
		for i = 1, GetNumTalentTabs() do
			local tabName = GetTalentTabInfo(i)
			if (tabName == BST["Demonology"]) then
				demonicKnowledgeTalentTab = i
				for j = 1, GetNumTalents(i) do
					local talentName, _, _, _, currentRank = GetTalentInfo(demonicKnowledgeTalentTab, j)
					if talentName == BS["Demonic Knowledge"] then
						demonicKnowledgeTalentIndex = j
						--set currentRank
                        demonicKnowledgeRank = currentRank
                        
					end
				end
			end
		end
	else
		local _, _, _, _, currentRank = GetTalentInfo(demonicKnowledgeTalentTab, demonicKnowledgeTalentIndex)
		-- set currentRank
        demonicKnowledgeRank = currentRank
	end
    self:Update()
end

function PetInfoFu:UpdateNetheredDemon(petName,petType,spellName,spellRank)
    if petName ~= nil then
        if self.db.char.netheredDemons[petName] == nil then
            self.db.char.netheredDemons[petName] = {}
        end
        self.db.char.netheredDemons[petName].PetName = petName
        self.db.char.netheredDemons[petName].PetType = petType
        if self.db.char.netheredDemons[petName][spellName] == nil then
            self.db.char.netheredDemons[petName][spellName] = {}
        end
        self.db.char.netheredDemons[petName][spellName] = {
            SpellName = spellName,
            SpellRank = spellRank
        }      
    end
end

function PetInfoFu:ToggleCategory(id, button)
    -- hidden is character specific data
    self.db.char.hidden[id] = not self.db.char.hidden[id]
	-- refresh in place
	self:UpdateTooltip()
end    
    
function PetInfoFu:PetInfo_GetPetExperience()
	local currentXP, totalXP = GetPetExperience()
	local toLevelXP = totalXP - currentXP
	return currentXP, totalXP, toLevelXP
end

function PetInfoFu:GetTimeToLevel(petXPToLevel)
    if timerZero == 0 then return L["NA"] end

    local petXPGained = petXPZero - petXPToLevel
    if petXPGained < 0 then 
        self:ResetXPTimer()
        petXPGained = petXPZero - petXPToLevel
    end

    local timeRemaining = {}
          timeRemaining.d = 0
          timeRemaining.h = 0
          timeRemaining.m = 0
          timeRemaining.s = 0
    local timeDelta = (petXPToLevel/petXPGained) * (time() - timerZero)
    
    -- 86,400 seconds equals one day
    if (timeDelta >= 86400) then
        timeRemaining.d = floor(timeDelta / 86400)
        timeDelta = (timeDelta - (timeRemaining.d * 86400))
    end
    if timeRemaining.d > 0 then return L["1D+"] end
    
    -- 3,600 seconds equals one hour
    if (timeDelta >= 3600) then
        timeRemaining.h = floor(timeDelta / 3600)
        timeDelta = (timeDelta - (timeRemaining.h * 3600))
    end
    -- 60 seconds equals one minute
    if ( timeDelta >= 60 ) then
        timeRemaining.m = floor(timeDelta / 60)
        timeDelta = (timeDelta - (timeRemaining.m * 60))
    end
    -- add remaining seconds
    timeRemaining.s = timeDelta
    
    return string.format(L["PI_TIME_FORMAT"], timeRemaining.h, timeRemaining.m)

end
    
function PetInfoFu:GetKillsToLevel(petXPToLevel)
    return lastKillXP..' ('..format(L["PI_FORMAT"], petXPToLevel/lastKillXP)..')'
end
    
-- is class functions
    
function PetInfoFu:IsHunter()
	local _, englishClass = UnitClass("player")
	return englishClass == "HUNTER"
end
	
function PetInfoFu:IsWarlock()
	local _, englishClass = UnitClass("player")
	return englishClass == "WARLOCK" 
end

-- play sound on higher level skill available

function PetInfoFu:CheckForNextLevelSkill()
    local petLevel = UnitLevel("pet") 
    if petLevel > prevPetLevel then
        if self.db.char.playSoundOnSkillTrainable and not self.db.char.hasPlayedSoundOnSkillTrainable then
            i = 1
            local petName = UnitName("pet")
            while true do
                local spellName, spellRank = GetSpellName(i, BOOKTYPE_PET)
                if not spellName then
                    do break end
                end
                -- check if higher skill is available
                if self.db.char.beastTraining[spellName] ~= nil then
                    if self.db.char.beastTraining[spellName][spellRank] ~= nil then
                        if string.find(spellRank,' ') then                   
                            local rankNo = string.sub(spellRank,string.find(spellRank,' ') + 1)
                            local nextRank = string.sub(spellRank,1,string.find(spellRank,' '))..rankNo + 1
                            if self.db.char.beastTraining[spellName][nextRank] ~= nil then
                                local requiredLevel = self.db.char.beastTraining[spellName][nextRank].RequiredLevel
                                if requiredLevel <= petLevel then 
                                    PlaySound('AuctionWindowOpen')                              
                                    self.db.char.hasPlayedSoundOnSkillTrainable = true
                                    DEFAULT_CHAT_FRAME:AddMessage(crayon:Green(L["NAME"]..': '..petName..L[" can learn a new skill."]))
                                end
                            end
                        end
                    end
                end
                i = i + 1
            end
        end
    end
    prevPetLevel = petLevel
end

function PetInfoFu:CheckForTPGain()
    local totalTP, usedTP = GetPetTrainingPoints()
    local freeTP = totalTP - usedTP
    
    if prevTP == 0 then prevTP = freeTP end
    -- play sound if pet has gained training points since last update
    if freeTP > prevTP then
        if self.db.char.playSoundOnGainTP and
           not self.db.char.hasPlayedSoundOnGainTP then
            PlaySound('AuctionWindowOpen')
            self.db.char.hasPlayedSoundOnGainTP = true
            local petName = UnitName("pet")
            DEFAULT_CHAT_FRAME:AddMessage(crayon:Green(L["NAME"]..': '..petName..L[" has gained Training Points."]));                       
        end
        prevTP = freeTP
    end
end	
    
-- map notes

function PetInfoFu:AddStableMapNote()
	--get coordinates and name of stable owner
	local x, y = GetPlayerMapPosition("player")
    local stableName, _ = UnitName("target")

    -- check if we have an existing note, if not add one
	if not Cartographer_Notes:GetNearbyNote(GetRealZoneText(), x, y, 0.05, 'PetInfoFu') then
        Cartographer_Notes:SetNote(GetRealZoneText(), x, y, L["Stable"], "PetInfoFu",'title',  stableName)
    end
end    
    
-- menus
	
function PetInfoFu:OnMenuRequest(level, value)
	if level == 1 then
		if self:IsHunter() then
            dewdrop:AddLine(
                'text', BS["Beast Training"],
                'func', function() CastSpellByName(BS["Beast Training"])
                        end,
                'closeWhenClicked', true
            )
            dewdrop:AddLine()
		end
        -- label options
        dewdrop:AddLine(
            'text', L["Show pet's name"],
    		'func', 'ToggleIsShowingPetName',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingPetName()
        )
        dewdrop:AddLine(
            'text', L["Show pet's level"],
    		'func', 'ToggleIsShowingPetLevel',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingPetLevel()
        )
        if self:IsHunter() then
            dewdrop:AddLine(
                'text', L["Show experience"],        -- stub for popup
                'arg1', self,
                'hasArrow', true,
                'value', "SHOWXP"
            )                
        end
        if self:IsHunter() then
            dewdrop:AddLine(
                'text', L["Play sound"],             -- stub for popup
                'arg1', self,
                'hasArrow', true,
                'value', "PLAYSOUND"
            )
        end            
        -- tooltip options
        dewdrop:AddLine(
            'text', L["Show pet's attributes"],
    		'func', 'ToggleIsShowingAttributes',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingAttributes()
        )
        dewdrop:AddLine(
            'text', L["Show pet's combat statistics"],
    		'func', 'ToggleIsShowingCombatStats',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingCombatStats()
        )
        dewdrop:AddLine(
            'text', L["Show pet's skills"],
    		'func', 'ToggleIsShowingSkills',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingSkills()
        )
       if self:IsHunter() then
            dewdrop:AddLine(
                'text', L["Track pet happiness"],
                'func', 'ToggleIsTrackingPetHappiness',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsTrackingPetHappiness()
            )
            dewdrop:AddLine(
                'text', L["CATEGORY_PET_STABLED"],  -- stub for popup
                'arg1', self,
                'hasArrow', true,
                'value', "STABLEDPETS"
            )        
            dewdrop:AddLine(
                'text', L["Track pet history"],
                'func', 'ToggleIsTrackingPetHistory',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsTrackingPetHistory()
            )
            dewdrop:AddLine(
                'text', L["MapNotes"],              -- stub for popup
                'arg1', self,
                'hasArrow', true,
                'value', "MAPNOTES"
            )        
        end
        if self:IsWarlock() then
            dewdrop:AddLine(
                'text', L["Show nethered demons"],
                'func', 'ToggleIsShowingStabledPets',
                'arg1', self,
                'closeWhenClicked', true,         
                'checked', self:IsShowingStabledPets()
            )        
        end        
        dewdrop:AddLine()
	end
    if level == 2 then
        if value == "STABLEDPETS" then
            dewdrop:AddLine(
                'text', L["Show stabled pets"],
                'func', 'ToggleIsShowingStabledPets',
                'arg1', self,
                'closeWhenClicked', true,         
                'checked', self:IsShowingStabledPets()
            )
            dewdrop:AddLine(
                'text', L["Show loyalty"],
                'func', 'ToggleIsShowingLoyalty',
                'arg1', self,
                'closeWhenClicked', true,         
                'checked', self:IsShowingLoyalty()
            )        
        end
        if value == "MAPNOTES" then
            dewdrop:AddLine(
                'text', L["Enable Stable mapnotes"],
                'func', 'ToggleIsShowingStableMapNotes',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsShowingStableMapNotes()
            )
        end
        if value == "SHOWXP" then
            dewdrop:AddLine(
                'text', L["By amount"],
                'func', 'ToggleIsShowingXPByAmount',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsShowingXPByAmount()
            )
            dewdrop:AddLine(
                'text', L["By kill"],
                'func', 'ToggleIsShowingXPByKill',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsShowingXPByKill()
            )
            dewdrop:AddLine(
                'text', L["By percentage"],
                'func', 'ToggleIsShowingXPByPercentage',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsShowingXPByPercentage()
            )
            dewdrop:AddLine(
                'text', L["By time"],
                'func', 'ToggleIsShowingXPByTime',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsShowingXPByTime()
            )
            dewdrop:AddLine(
                'text', L["Reset timer"],
                'func', 'ResetXPTimer',
                'arg1', self,
                'closeWhenClicked', true
            )
            dewdrop:AddLine(
                'text', L["Show on bar"],
                'func', 'ToggleIsShowingXPOnBar',
                'arg1', self,
                'closeWhenClicked', true
            )
        end 
        if value == "PLAYSOUND" then
            dewdrop:AddLine(
                'text', L["On TP gain"],
                'func', 'TogglePlaySoundOnGainTP',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsPlayingSoundOnGainTP()
            )
            dewdrop:AddLine(
                'text', L["On new skill available"],
                'func', 'TogglePlaySoundOnSkillTrainable',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsPlayingSoundOnSkillTrainable()
            )
        end       
    end
end
	
-- updates    

function PetInfoFu:AddStabledPets()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_STABLED"], 'columns', 2,
                                           'text', L["CATEGORY_PET_STABLED"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_STABLED"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_STABLED"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_STABLED"]] then
        local petsFound
        
        for k, v in pairs(self.db.char.stabledPets) do  
            local label = v.PetLevel
            if self.db.char.showLoyalty then label = label..' ('..v.PetLoyalty..')' end 
            tooltipLine:AddLine(
                'text', v.PetName..' ('..v.PetType..')',
                'text2', label,
                'text2R', 1, 'text2G', 1, 'text2B', 0,
                'func', 'PetInfoFrame', 'arg1', self, 'arg2', v
            )              
            petsFound = true
        end
        if not petsFound then
                tooltipLine:AddLine(
                    'text', L["Please visit a stable to initialise."]
                )
        end
    end
end
	
function PetInfoFu:AddPetHistory()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_HISTORY"], 'columns', 2,
                                           'text', L["CATEGORY_PET_HISTORY"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_HISTORY"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_HISTORY"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_HISTORY"]] then
        -- create a sorted table of pets
        sortedPets = {}
        for k,v in pairs(self.db.char.petHistory) do
            if v.HunterLevel == nil then
                v.HunterLevel = UnitLevel("player")
                self.db.char.petHistory[k].HunterLevel = v.HunterLevel
            end
            table.insert(sortedPets, v)
        end
        table.sort(sortedPets, function (a, b) return a.HunterLevel < b.HunterLevel end)
        -- list sorted pets
        for i = 1, #sortedPets do  
            tooltipLine:AddLine(
                'text', sortedPets[i].PetName..' ('..sortedPets[i].PetType..')',
                'text2', sortedPets[i].PetLevel..' ('..sortedPets[i].HunterLevel..')',
                'text2R', 1, 'text2G', 1, 'text2B', 0
            )              
        end
    end
end


function PetInfoFu:ShowSkills()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_SKILLS"], 'columns', 2,
                                           'text', L["CATEGORY_PET_SKILLS"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_SKILLS"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_SKILLS"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_SKILLS"]] then
        local petName = UnitName("pet")    
        local petType = UnitCreatureFamily("pet")
        local petLevel = UnitLevel("pet") 
        local i = 1
        while true do
            local spellName, spellRank = GetSpellName(i, BOOKTYPE_PET)
            if not spellName then
                do break end
            end
            -- check if higher skill is available
            if self:IsHunter() then
                if self.db.char.beastTraining[spellName] ~= nil then
                    if self.db.char.beastTraining[spellName][spellRank] ~= nil then
                        if string.find(spellRank,' ') then
                            local rankNo = string.sub(spellRank,string.find(spellRank,' ') + 1)
                            local nextRank = string.sub(spellRank,1,string.find(spellRank,' '))..rankNo + 1
                            if self.db.char.beastTraining[spellName][nextRank] ~= nil then
                                local requiredLevel = self.db.char.beastTraining[spellName][nextRank].RequiredLevel
                                if requiredLevel > petLevel then 
                                    spellRank = spellRank..' ('..crayon:Red(format(requiredLevel))..')'
                                else
                                    spellRank = spellRank..' ('..format(requiredLevel)..')'
                                end
                            end
                        end
                    end
                end
            end
            if self:IsWarlock() then
                self:UpdateNetheredDemon(petName,petType,spellName,spellRank)
            end
            tooltipLine:AddLine(
                'text', spellName,
                'text2', spellRank,
                'text2R', 1, 'text2G', 1, 'text2B', 0
            )
            i = i + 1
        end
     end
end     

function PetInfoFu:AddNetheredDemons()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_NETHER"], 'columns', 2,
                                           'text', L["CATEGORY_PET_NETHER"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_NETHER"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                              			   'child_indentation', 20,                                           
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_NETHER"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    
    if not self.db.char.hidden[L["CATEGORY_PET_NETHER"]] then
        -- create a sorted table of pets
        local sortedPets = {}
        for k,v in pairs(self.db.char.netheredDemons) do
             table.insert(sortedPets, v)
        end
        table.sort(sortedPets, function (a, b) return a.PetType < b.PetType end)
        -- list sorted pets
        for i = 1, #sortedPets do
            if sortedPets[i].PetName ~= UnitName('pet') then
                local tooltipLine = tablet:AddCategory('id', sortedPets[i].PetName, 'columns', 2,
                                                       'text', sortedPets[i].PetName..' ('..sortedPets[i].PetType..')',
                                                       'func', 'ToggleCategory', 'arg1', self, 'arg2', sortedPets[i].PetName,
                                                       'indentation', 10,
                                                       'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                       'child_indentation', 25,
                                                       'showWithoutChildren', true,
                                                       'hideBlankLine', true,                                           
                                                       'checked', true, 'hasCheck', true, 'checkIcon',
                                                       self.db.char.hidden[sortedPets[i].PetName] and 
                                                       'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
                )        
                if not self.db.char.hidden[sortedPets[i].PetName] then
                    -- create a sorted table of skills
                    local sortedSkills = {}
                    for k,v in pairs(self.db.char.netheredDemons[sortedPets[i].PetName]) do
                        if v.SpellName ~= nil then 
                            table.insert(sortedSkills, v)
                        end
                    end
                    table.sort(sortedSkills, function (a, b) return a.SpellName < b.SpellName end)
                    -- list sorted skills
                    for j = 1, #sortedSkills do
                        tooltipLine:AddLine(
                            'text', sortedSkills[j].SpellName,
                            'text2', sortedSkills[j].SpellRank,
                            'text2R', 1, 'text2G', 1, 'text2B', 0
                        ) 
                    end                    
                end
            end
        end
    end
end
 
function PetInfoFu:OnTooltipUpdate()

    if HasPetUI() then

        if self:IsHunter() then
	        tablet:SetHint(L["PI_TOOLTIP_HINT_HUNTER"])
        else
            tablet:SetHint(L["PI_TOOLTIP_HINT"])       
        end

        
        local toolTip = tablet:AddCategory(
            'columns', 2,
            'child_textR', 0,
            'child_textG', 1,
            'child_textB', 0,
            'child_text2R', 1,
            'child_text2G', 1,
            'child_text2B', 0
        )
        
        local petName = UnitName("pet")
        toolTip:AddLine(
            'text', L["PI_PETNAME"],
            'text2', petName
        )        
        local petLevel = UnitLevel("pet")
        toolTip:AddLine(
            'text', L["PI_PETLEVEL"],
            'text2', petLevel
        )
        if self:IsHunter() then        
            -- general pet info            
            local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_INFO"], 'columns', 2,
                                                   'text', L["CATEGORY_PET_INFO"],
                                                   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_INFO"],
                                                   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                   'showWithoutChildren', true,
                                                   'checked', true, 'hasCheck', true, 'checkIcon',
                                                   self.db.char.hidden[L["CATEGORY_PET_INFO"]] and 
                                                   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
            )
            if not self.db.char.hidden[L["CATEGORY_PET_INFO"]] then
            
                local petType = UnitCreatureFamily("pet")
                
                tooltipLine:AddLine(
                    'text', L["PI_TYPE"],
                    'text2', petType,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine(
                    'text', L["PI_FOOD"],
                    'text2', BuildListString(GetPetFoodTypes()),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine(
                    'text', L["PI_LOYALTY"],
                    'text2', GetPetLoyalty(),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()
                
                -- happiness info
                local happiness, damagePercentage, loyaltyRate = GetPetHappiness()
                local happyText = ''
                local loyaltyText = ''
                local rHappy = 1 
                local gHappy = 1
                local bHappy = 0

                if happiness == 1 then
                    happyText = L["Unhappy"]
                    rHappy = 1
                    gHappy = 0
                elseif happiness == 2 then
                    happyText = L["Content"]              
                elseif happiness == 3 then
                    happyText = L["Happy"]
                    rHappy = 0
                    gHappy = 1
                end
                
                if loyaltyRate > 0 then
                    loyaltyText = L["Gaining Loyalty"]
                else
                    loyaltyText = L["Losing Loyalty"]
                end
                tooltipLine:AddLine(
                    'text', L["PI_HAPPINESS"],
                    'text2', happyText,
                    'text2R', rHappy, 'text2G', gHappy, 'text2B', bHappy
                )            
                tooltipLine:AddLine(
                    'text', L["PI_DAMAGEPERCT"],
                    'text2', damagePercentage..'%',
                    'text2R', rHappy, 'text2G', gHappy, 'text2B', bHappy
                )
                tooltipLine:AddLine(
                    'text', L["PI_LOYALTYRATE"],
                    'text2', loyaltyText,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()
                
                -- experience info          
                local currentXP, totalXP, toLevelXP = self:PetInfo_GetPetExperience()
                local currentXPPercent = currentXP / totalXP * 100
                local toLevelXPPercent = toLevelXP / totalXP * 100
                
                if petLevel < UnitLevel("player") then
                    tooltipLine:AddLine(
                        'text', L["PI_CURRENT_XP"], 
                        'text2', format(L["PI_PLUS_PERCENT_FORMAT"], currentXP, currentXPPercent),
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                    tooltipLine:AddLine(
                        'text', L["PI_NEEDED_XP"], 
                        'text2', format(L["PI_PLUS_PERCENT_FORMAT"], toLevelXP, toLevelXPPercent),
                        'text2R', 1, 'text2G', 1, 'text2B', 0,
                        'func', 'PetExperienceFrame', 'arg1', self
                    )
                    tooltipLine:AddLine(
                        'text', L["PI_TOTAL_XP"], 
                        'text2', totalXP,
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                else
                    tooltipLine:AddLine(
                        'text', L["PI_NOLEVEL"]
                    )
                end
                tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()
                
                -- training info
                local totalTP, usedTP = GetPetTrainingPoints()
                local freeTP = totalTP - usedTP            
                tooltipLine:AddLine(
                    'text', L["PI_TRAINING_POINTS"], 
                    'text2', freeTP,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
            end
        end
        
        if self:IsWarlock() then
            -- demonic knowledge
            -- skip section if not talented in demonic knowledge
            if demonicKnowledgeRank > 0 then
                local tooltipLine = tablet:AddCategory('id', L["CATEGORY_DEMONIC_KNOWLEDGE"], 'columns', 2,
                                                       'text', L["CATEGORY_DEMONIC_KNOWLEDGE"],
                                                       'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_DEMONIC_KNOWLEDGE"],                                                   
                                                       'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                       'showWithoutChildren', true,
                                                       'checked', true, 'hasCheck', true, 'checkIcon',
                                                       self.db.char.hidden[L["CATEGORY_DEMONIC_KNOWLEDGE"]] and 
                                                       'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'                                                  
                )
            
                if UnitExists("pet") then
                    local demonSta, demonInt
                    _, demonSta = UnitStat("pet", 3) -- stamina
                    _, demonInt = UnitStat("pet", 4) -- intelligence
                    tooltipLine:AddLine(
                        'text', L["PI_BONUSDAMAGE"]..' ('..demonicKnowledgeRank..')',
                        'text2', floor((demonSta + demonInt) * 0.05 * demonicKnowledgeRank),
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                else
                    tooltipLine:AddLine(
                        'text', L["PI_NODEMON"]
                    )            
                end
            end
        end
        
        -- combat info
        if self.db.char.showCombatStats then        
            local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_COMBAT_INFO"], 'columns', 2,
                                           'text', L["CATEGORY_PET_COMBAT_INFO"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_COMBAT_INFO"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_COMBAT_INFO"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
            )
            if not self.db.char.hidden[L["CATEGORY_PET_COMBAT_INFO"]] then

                local baseArmor, effectiveArmor, armor, posArmorBuff, negArmorBuff = UnitArmor("pet");
                if posArmorBuff - negArmorBuff < 0 then
                    effectiveArmor = crayon:Red(format(effectiveArmor))..' ('..format(baseArmor)..crayon:Red("-"..format(baseArmor - effectiveArmor))..')'
                elseif baseArmor < effectiveArmor then 
                    effectiveArmor = crayon:Green(format(effectiveArmor))..' ('..format(baseArmor)..crayon:Green("+"..format(effectiveArmor - baseArmor))..')'
                end
-- DEFAULT_CHAT_FRAME:AddMessage(baseArmor.." "..effectiveArmor.." "..armor.." "..posArmorBuff.." ".." "..negArmorBuff)                

                tooltipLine:AddLine(
                    'text', L["PI_ARMOR_CLASS"], 
                    'text2', effectiveArmor,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
              
                local atkBase, atkMod = UnitAttackBothHands("pet");        
                tooltipLine:AddLine(
                    'text', L["PI_ATTACK_SKILL"],
                    'text2', atkBase,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )

                local atkPowerbase, atkPowerPosBuff, atkPowerNegBuff = UnitAttackPower("pet");
                if atkPowerPosBuff - atkPowerNegBuff < 0 then
                    atkPowerbase = crayon:Red(format(atkPowerbase + atkPowerPosBuff + atkPowerNegBuff))..' ('..format(atkPowerbase)..crayon:Red("-"..format(atkPowerPosBuff + atkPowerNegBuff))..')'
                elseif atkPowerPosBuff - atkPowerNegBuff > 0 then
                    atkPowerbase = crayon:Green(format(atkPowerbase + atkPowerPosBuff + atkPowerNegBuff))..' ('..format(atkPowerbase)..crayon:Green("+"..format(atkPowerPosBuff + atkPowerNegBuff))..')'
                end
                tooltipLine:AddLine(
                    'text', L["PI_ATTACK_POWER"],
                    'text2', atkPowerbase,
                    'text2R', 1, 'text2G', 1, 'text2B', 0                
                )
                
                local atkSpeed = UnitAttackSpeed("pet")       
                tooltipLine:AddLine(
                    'text', L["PI_ATTACK_SPEED"], 
                    'text2', format(L["PI_AS_FORMAT"], atkSpeed),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )

                local lowDmg, hiDmg, offlowDmg, offhiDmg, posBuff, negBuff, percentMod = UnitDamage("pet");
                local listedDPS = (((lowDmg + hiDmg) * .5 + posBuff + negBuff) * percentMod) / atkSpeed;
                tooltipLine:AddLine(
                    'text', L["PI_DPS_LISTED"], 
                    'text2', format(L["PI_DPS_FORMAT"], listedDPS),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
            end
        end
        
        -- show stats info if showing
        if self.db.char.showAttributes then        
            local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_ATTRIBUTES"], 'columns', 2,
                                                   'text', L["CATEGORY_PET_ATTRIBUTES"],
                                                   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_ATTRIBUTES"],
                                                   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                   'showWithoutChildren', true,
                                                   'checked', true, 'hasCheck', true, 'checkIcon',
                                                   self.db.char.hidden[L["CATEGORY_PET_ATTRIBUTES"]] and 
                                                   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
            )        
            if not self.db.char.hidden[L["CATEGORY_PET_ATTRIBUTES"]] then         
                for i = 1, NUM_PET_STATS, 1 do                  
                    local attrbStr = L["SPELL_STAT"..format(i-1).."_NAME"]                   
                    local baseStat, effectiveStat, posBuff, negBuff = UnitStat("pet", i)
            
                    if posBuff == 0 and negBuff == 0 then
                        baseStat = effectiveStat
                    else 
                        -- if there are any negative buffs then show the main number in red even if there are
                        -- positive buffs. otherwise show in green.
                        if negBuff < 0 then
                            baseStat = crayon:Red(effectivestat)..' ('..baseStat.."-"..crayon:Red(negBuff+posBuff)..')'
                        else
                            baseStat = crayon:Green(effectiveStat)..' ('..baseStat-posBuff.."+"..crayon:Green(posBuff)..')'
                        end
                    end
                    tooltipLine:AddLine(
                        'text', attrbStr,
                        'text2', baseStat,
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                end
            end
        end
        
        -- show skills info
        if self:IsHunter() and self.db.char.showSkills then          
            self:ShowSkills()
        end

        -- add other demons if showing
        if self:IsWarlock() and self.db.char.showStabledPets then          
            self:AddNetheredDemons()
        end   
        
        -- add other pets if showing
        if self:IsHunter() and self.db.char.showStabledPets then          
            self:AddStabledPets()
        end   
        -- add pet history
        if self:IsHunter() and self.db.char.trackPetHistory then
            self:AddPetHistory()
        end           
    else
        -- no pet found
        local toolTip = tablet:AddCategory(
            'columns', 1,
            'child_textR', 0,
            'child_textG', 1,
            'child_textB', 0
        )      
        toolTip:AddLine(
            'text', L["PI_NOPET"]
        )
        -- add stabled pets
        if self:IsHunter() and self.db.char.showStabledPets then
            self:AddStabledPets()
        end   
        -- add pet history
        if self:IsHunter() and self.db.char.trackPetHistory then
            self:AddPetHistory()
        end    
        -- add other demons if showing
        if self:IsWarlock() and self.db.char.showStabledPets then          
            self:AddNetheredDemons()
        end   
    end
end

function PetInfoFu:UpdateText()
    -- sound checks
    if self:IsHunter() then
        self:CheckForTPGain()
        self:CheckForNextLevelSkill() 
    end

    if  self:IsTextShown() then
        if self.db.char.showPetName or self.db.char.showPetLevel then
            if UnitName("pet") ~= nil then
                local petName = UnitName("pet")
                local petLevel = UnitLevel("pet")
                local toonLevel = UnitLevel("player")
                local petIcon = GetPetIcon()
                local happiness = GetPetHappiness()
                local label = ''
                
                if petName ~= nil and petLevel ~= nil then  
                
                    if petName == UNKNOWNOBJECT then
                        petName = L["UO"]
                        self:ScheduleEvent(self.UpdateText,2, self)
                    end
                
                    -- set pet icon for warlock
                    if self:IsWarlock() then
                        if (UnitCreatureFamily("pet") == L["Imp"]) then            petIcon = BS:GetSpellIcon(BS["Summon Imp"]) -- "Interface\\Icons\\Spell_Shadow_SummonImp"
                        elseif (UnitCreatureFamily("pet") == L["Felguard"]) then   petIcon = BS:GetSpellIcon(BS["Summon Felguard"]) -- "Interface\\Icons\\Spell_Shadow_SummonFelGuard"
                        elseif (UnitCreatureFamily("pet") == L["Felhunter"]) then  petIcon = BS:GetSpellIcon(BS["Summon Felhunter"])-- "Interface\\Icons\\Spell_Shadow_SummonFelHunter"
                        elseif (UnitCreatureFamily("pet") == L["Succubus"]) then   petIcon = BS:GetSpellIcon(BS["Summon Succubus"]) -- "Interface\\Icons\\Spell_Shadow_SummonSuccubus"
                        elseif (UnitCreatureFamily("pet") == L["Voidwalker"]) then petIcon = BS:GetSpellIcon(BS["Summon Voidwalker"])-- "Interface\\Icons\\Spell_Shadow_SummonVoidWalker"
                        elseif (UnitCreatureFamily("pet") == L["Infernal"]) then   petIcon = "Interface\\Icons\\Spell_Shadow_SummonInfernal" 
                        end
                    end 
                    
                    if petIcon ~= nil then
                        self:SetIcon(petIcon)        
                    end
                    -- show name and/or level
                    if self.db.char.showPetName then                   
                        if self.db.char.showPetLevel then                        
                            label = petName..' ('..petLevel..')'
                        else
                            label = petName
                        end
                    else
                        if self.db.char.showPetLevel then    
                            label = petLevel
                        else
                                                
                        end 
                    end
                    
                    -- show XP if flag is true (otherwise assume XP tool tip is showing)
                    if self.db.char.showPetXPOnBar  then
                        if petLevel ~= toonLevel then
                            if self.db.char.showPetXP == 1 then
                                local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
                                label = label..' - '..format(L["PI_FORMAT"], toLevelXP)
                            elseif self.db.char.showPetXP == 2 then
                                local _, totalXP, toLevelXP = self:PetInfo_GetPetExperience()
                                local toLevelXPPercent = toLevelXP / totalXP * 100                   
                                label = label..' - '..format(L["PI_PERCENT_FORMAT"],toLevelXPPercent)
                            elseif self.db.char.showPetXP == 3 then
                                local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
                                label = label..' - '..self:GetTimeToLevel(toLevelXP)
                            elseif self.db.char.showPetXP == 4 then
                                if lastKillXP > 0 then 
                                    local _, _, toLevelXP = self:PetInfo_GetPetExperience()
                                    label = label..' - '..self:GetKillsToLevel(toLevelXP)
                                end
                            end
                        end
                    end
                    
                    -- set colour
                    if happiness == nil then                -- non hunter pet
                        self:SetText(label)
                    else
                        if self.db.char.trackPetHappiness then
                            label = label..format(L["PI_FORMAT_HAPPINESS"],estimatedPetHappiness) 
                        end
                        if happiness == 1 then
                            self:SetText(crayon:Red(label))     -- unhappy
                        elseif happiness == 2 then
                            self:SetText(crayon:Yellow(label))  -- content              
                        elseif happiness == 3 then
                            self:SetText(crayon:Green(label))   -- happy
                        end
                    end
                else
                    -- pet name and level are nil which means there shouldn't have been a pet ui in the first place
                    self:SetText(L["PI_NOPET"])
                    self:SetIcon(BS:GetSpellIcon("Train Beast"))
                end
            else 
                -- no pet ui found
                self:SetText(L["PI_NOPET"])
                self:SetIcon(BS:GetSpellIcon("Train Beast"))
            end
        else
            if UnitName("pet") ~= nil then
                local petIcon = GetPetIcon()
                if petIcon ~= nil then
                    self:SetIcon(petIcon)        
                end
            end
            self:SetText("")                    
        end
    end
end	

-- event handlers

function PetInfoFu:OnClick()
    if self:IsHunter() then
        if IsControlKeyDown() then
            -- initialise timer for show time til next level
            if self.db.char.showPetXP == 3 then
                self:ResetXPTimer()
            -- initialise XP counter for show kills til next level
            elseif self.db.char.showPetXP == 4 then
                local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
                prevXPToLevel = toLevelXP
            end            
        else
            self.db.char.hasPlayedSoundOnGainTP = false
            self.db.char.hasPlayedSoundOnSkillTrainable = false
            CastSpellByName(BS["Beast Training"])
        end
    end
end
